from PySide6.QtWidgets import QDialog, QGridLayout, QGroupBox, QButtonGroup, QVBoxLayout, QRadioButton, QCheckBox, QLabel, QPushButton, QLineEdit, QDoubleSpinBox, QMessageBox
from PySide6.QtCore import Qt
from PySide6.QtGui import QDoubleValidator
import numpy as np
from takefits.ui.save_fits_dialog import SaveFITS
from takefits.core import usecases
from takefits.core.history_provenance import build_processing_history_lines
from takefits.tools.base_panel import clear_action_preview_record, confirm_pending_close, has_action_record_tag, record_action_preview
import time
import os

class SmoothSettingsPanel(QDialog):
    def __init__(self, fits_viewer, subwindows):
        super().__init__()
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose, True)
        self.fits_viewer = fits_viewer
        self.header = self.fits_viewer.header
        self.wcs = self.fits_viewer.wcs
        self.subwindows = subwindows
        
        self.data = self.fits_viewer.data
        self.original_data = None  # Defer copying until first execution
        self._has_pending_changes = False
        self._action_record_tag = "panel:smoothing"
        self._last_action_name = None
        self._last_action_params = None
        
        self.initUI()
        

        self.bmaj_exists = 'BMAJ' in self.header
        self.bmin_exists = 'BMIN' in self.header
        self.bpa_exists = 'BPA' in self.header
            
        if self.bmaj_exists:
            self.original_bmaj = self.header['BMAJ']
            self.bmaj_in_arcsec = self.header['BMAJ'] * 3600
            self.bmaj_current.setText(f'{self.bmaj_in_arcsec:.3g}')
        if self.bmin_exists:
            self.original_bmin = self.header['BMIN']
            self.bmin_in_arcsec = self.header['BMIN'] * 3600
            self.bmin_current.setText(f'{self.bmin_in_arcsec:.3g}')
        if self.bpa_exists:
            self.original_bpa = self.header['BPA']
            self.bpa_in_deg = self.header['BPA']
            self.bpa_current.setText(f'{self.bpa_in_deg:.3g}')
        
        self.kernel_selection_changed()
        self.resync_after_workspace_restore()
        
    def initUI(self):    
        grid_layout = QGridLayout()
        grid_layout.setHorizontalSpacing(10)
        grid_layout.setVerticalSpacing(10)
        grid_layout.setContentsMargins(15, 0, 10, 13)
        
        self.smoothness_groupbox = QGroupBox("Smoothness")
        smoothness_layout = QGridLayout()

        smoothness_layout.addWidget(QLabel("X"), 0, 0)
        self.x_spinbox = QDoubleSpinBox()
        self.x_spinbox.setValue(1.00)
        self.x_spinbox.setSingleStep(1.)
        smoothness_layout.addWidget(self.x_spinbox, 0, 1)

        smoothness_layout.addWidget(QLabel("Y"), 1, 0)
        self.y_spinbox = QDoubleSpinBox()
        self.y_spinbox.setValue(1.00)
        self.y_spinbox.setSingleStep(1.)
        smoothness_layout.addWidget(self.y_spinbox, 1, 1)
        
        if  self.subwindows:
            smoothness_layout.addWidget(QLabel("Z"), 2, 0)
            self.z_spinbox = QDoubleSpinBox()
            self.z_spinbox.setValue(0.00)
            self.z_spinbox.setSingleStep(1.)
            smoothness_layout.addWidget(self.z_spinbox, 2, 1)

        self.smoothness_groupbox.setLayout(smoothness_layout)

        grid_layout.addWidget(self.smoothness_groupbox, 0, 0, 4, 1)

        kernel_groupbox = QGroupBox("Kernel")
        kernel_layout = QVBoxLayout()

        self.gaussian_radio = QRadioButton("Gaussian")
        self.boxcar_radio = QRadioButton("Boxcar")
        self.hanning_radio = QRadioButton("Hanning")
        self.gaussian_radio.setChecked(True)
        self.hanning_radio.setEnabled(self.data.ndim == 3)
        self.hanning_radio.setToolTip("Available only for 3D cubes")

        self.kernel_button_group = QButtonGroup()
        self.kernel_button_group.addButton(self.gaussian_radio)
        self.kernel_button_group.addButton(self.boxcar_radio)
        self.kernel_button_group.addButton(self.hanning_radio)

        kernel_layout.addWidget(self.gaussian_radio)
        kernel_layout.addWidget(self.boxcar_radio)
        kernel_layout.addWidget(self.hanning_radio)
        kernel_groupbox.setLayout(kernel_layout)

        grid_layout.addWidget(kernel_groupbox, 0, 1, 3, 1)

        self.checkbox = QCheckBox("Target res.")
        grid_layout.addWidget(self.checkbox, 3, 1, alignment=Qt.AlignmentFlag.AlignCenter)

        self.target_res_groupbox = QGroupBox("Target resolution")
        target_res_layout = QGridLayout()
        target_res_layout.setSpacing(2)

        target_res_layout.addWidget(QLabel("BMAJ"), 0, 0)
        self.bmaj_target = QLineEdit()
        self.bmaj_target.setFixedWidth(50)
        target_res_layout.addWidget(self.bmaj_target, 0, 1)
        target_res_layout.addWidget(QLabel("arcsec"), 0, 2)
        self.bmaj_target.setValidator(QDoubleValidator())
        
        target_res_layout.addWidget(QLabel("BMIN"), 1, 0)
        self.bmin_target = QLineEdit()
        self.bmin_target.setFixedWidth(50)
        target_res_layout.addWidget(self.bmin_target, 1, 1)
        target_res_layout.addWidget(QLabel("arcsec"), 1, 2)
        self.bmin_target.setValidator(QDoubleValidator())

        target_res_layout.addWidget(QLabel("BPA"), 2, 0)
        self.bpa_target = QLineEdit()
        self.bpa_target.setFixedWidth(50)
        target_res_layout.addWidget(self.bpa_target, 2, 1)
        target_res_layout.addWidget(QLabel("deg"), 2, 2)
        self.bpa_target.setValidator(QDoubleValidator())

        self.target_res_groupbox.setLayout(target_res_layout)
        grid_layout.addWidget(self.target_res_groupbox, 0, 2, 4, 1)

        current_res_groupbox = QGroupBox("Current resolution")
        current_res_layout = QGridLayout()
        current_res_layout.setSpacing(2)

        current_res_layout.addWidget(QLabel("BMAJ"), 0, 0)
        self.bmaj_current = QLineEdit()
        self.bmaj_current.setFixedWidth(50)
        self.bmaj_current.setReadOnly(True)
        current_res_layout.addWidget(self.bmaj_current, 0, 1)
        current_res_layout.addWidget(QLabel("arcsec"), 0, 2)

        current_res_layout.addWidget(QLabel("BMIN"), 1, 0)
        self.bmin_current = QLineEdit()
        self.bmin_current.setFixedWidth(50)
        self.bmin_current.setReadOnly(True)
        current_res_layout.addWidget(self.bmin_current, 1, 1)
        current_res_layout.addWidget(QLabel("arcsec"), 1, 2)

        current_res_layout.addWidget(QLabel("BPA"), 2, 0)
        self.bpa_current = QLineEdit()
        self.bpa_current.setFixedWidth(50)
        self.bpa_current.setReadOnly(True)
        current_res_layout.addWidget(self.bpa_current, 2, 1)
        current_res_layout.addWidget(QLabel("deg"), 2, 2)

        current_res_groupbox.setLayout(current_res_layout)
        grid_layout.addWidget(current_res_groupbox, 0, 3, 4, 1)

        self.execute_button = QPushButton("Execute")
        self.execute_button.clicked.connect(self.execute_smoothing)
        grid_layout.addWidget(self.execute_button, 4, 0, 1, 1) 
        self.execute_button.setAutoDefault(True) 
        self.execute_button.setDefault(True)

        self.reset_button = QPushButton("Reset")
        self.reset_button.clicked.connect(self.reset_smoothing)
        grid_layout.addWidget(self.reset_button, 4, 1, 1, 1) 
        

        self.save_button = QPushButton("Save as FITS")
        self.save_button.clicked.connect(self.save_fits)
        grid_layout.addWidget(self.save_button, 4, 2, 1, 1) 


        self.setLayout(grid_layout)

        self.checkbox.stateChanged.connect(self.toggle_target_res_group)
        self.checkbox.setChecked(False)
        self.target_res_groupbox.setEnabled(False)

        self.smoothness_groupbox.setEnabled(True)

        self.gaussian_radio.toggled.connect(self.kernel_selection_changed)
        self.boxcar_radio.toggled.connect(self.kernel_selection_changed)
        self.hanning_radio.toggled.connect(self.kernel_selection_changed)

        self.checkbox.setEnabled(True)
        
        self.setWindowTitle(f'Smooth Panel: {self.fits_viewer.filename}')
        self.move_to_default_position()

    def toggle_target_res_group(self, state=None):
        is_checked = self.checkbox.isChecked()
        self.target_res_groupbox.setEnabled(is_checked)
        self.bmaj_target.setFocus()
        self.smoothness_groupbox.setEnabled(not is_checked)
        
        if is_checked:
            if not self.bmaj_exists:
                self.bmaj_current.setReadOnly(False)
            if not self.bmin_exists:
                self.bmin_current.setReadOnly(False)
            if not self.bpa_exists:
                self.bpa_current.setReadOnly(False)
        else:
            self.bmaj_current.setReadOnly(True)
            self.bmin_current.setReadOnly(True)
            self.bpa_current.setReadOnly(True)

    def kernel_selection_changed(self, _checked=None):
        if self.gaussian_radio.isChecked():
            self.checkbox.setEnabled(True)
            self.toggle_target_res_group()
            return

        self.checkbox.setChecked(False)
        self.checkbox.setEnabled(False)
        self.target_res_groupbox.setEnabled(False)
        self.bmaj_current.setReadOnly(True)
        self.bmin_current.setReadOnly(True)
        self.bpa_current.setReadOnly(True)

        if self.boxcar_radio.isChecked():
            self.smoothness_groupbox.setEnabled(True)
        elif self.hanning_radio.isChecked():
            self.smoothness_groupbox.setEnabled(False)
        else:
            self.smoothness_groupbox.setEnabled(True)

    def execute_smoothing(self):
        if self.original_data is None:
            self.original_data = self.fits_viewer.data.copy()

        # Determine kernel type
        if self.gaussian_radio.isChecked():
            kernel_type = 'gaussian'
        elif self.boxcar_radio.isChecked():
            kernel_type = 'boxcar'
        else:
            kernel_type = 'hanning'

        if self.checkbox.isChecked() and kernel_type == 'gaussian':
            # Target resolution smoothing
            smoothed_data = self._execute_target_resolution_smoothing()
            if smoothed_data is None:
                return
        else:
            # Normal smoothing with pixel-based kernel size
            smoothed_data = self._execute_normal_smoothing(kernel_type)
            if smoothed_data is None:
                return

        # Update data in viewer and subwindows
        self.fits_viewer.data = smoothed_data
        self.fits_viewer.update_cube()

        for window in self.subwindows:
            if window:
                window.data = self.fits_viewer.data
                window.update_cube()

        self._has_pending_changes = True
        self._sync_app_state_data()
        self.update_all_displays()
        self._refresh_main_hpbw_overlay()
        self._record_preview_action()

    def _execute_normal_smoothing(self, kernel_type: str):
        """Execute normal smoothing using usecase layer."""
        self._last_action_name = "apply_smoothing"
        smoothness_x = self.x_spinbox.value()
        smoothness_y = self.y_spinbox.value()
        smoothness_z = self.z_spinbox.value() if self.subwindows else 0

        if kernel_type == 'hanning':
            if self.original_data.ndim != 3:
                QMessageBox.warning(self, 'Unsupported Data', 'Hanning is available only for 3D cubes.')
                return None
            self._last_action_params = {
                "kernel_type": kernel_type,
            }
            print("\n\nHanning kernel (velocity axis): [0.25, 0.5, 0.25] channels")
        else:
            self._last_action_params = {
                "kernel_type": kernel_type,
                "smoothness_x": float(smoothness_x),
                "smoothness_y": float(smoothness_y),
                "smoothness_z": float(smoothness_z),
            }

            if smoothness_x == 0 and smoothness_y == 0 and smoothness_z == 0:
                self.reset_smoothing()
                return None

            # Log kernel info
            if kernel_type == 'gaussian':
                fwhm_factor = 2.0 * np.sqrt(2.0 * np.log(2))
                print(f"\n\nGaussian kernel size (FWHM): ({fwhm_factor*smoothness_x:.3g}, {fwhm_factor*smoothness_y:.3g}, {fwhm_factor*smoothness_z:.3g}) pixels")
            else:
                size_x = int(2 * smoothness_x + 1) if smoothness_x > 0 else 1
                size_y = int(2 * smoothness_y + 1) if smoothness_y > 0 else 1
                size_z = int(2 * smoothness_z + 1) if smoothness_z > 0 else 1
                print(f"\n\nBoxcar kernel size: ({size_x}, {size_y}, {size_z}) pixels")

        # Call usecase layer
        current_time = time.time()
        print("Convolving kernel...")

        # Handle 4D data by applying to each slice
        data = self.original_data
        if data.ndim == 4:
            smoothed_data = np.empty_like(data)
            for i in range(data.shape[0]):
                smoothed_data[i] = usecases.compute_smoothed(
                    data[i],
                    kernel_type=kernel_type,
                    smoothness_x=smoothness_x,
                    smoothness_y=smoothness_y,
                    smoothness_z=smoothness_z,
                    handle_nan=True,
                )
        else:
            smoothed_data = usecases.compute_smoothed(
                data,
                kernel_type=kernel_type,
                smoothness_x=smoothness_x,
                smoothness_y=smoothness_y,
                smoothness_z=smoothness_z,
                handle_nan=True,
            )

        elapsed_time = time.time() - current_time
        print(f"Smoothing done in {elapsed_time:.3g} sec")
        return smoothed_data

    def _execute_target_resolution_smoothing(self):
        """Execute target resolution smoothing using usecase layer."""
        try:
            current_bmaj = float(self.bmaj_current.text())  # in arcsec
            current_bmin = float(self.bmin_current.text())
            current_bpa = float(self.bpa_current.text())  # in degrees
            target_bmaj = float(self.bmaj_target.text())
            target_bmin = float(self.bmin_target.text())
            target_bpa = float(self.bpa_target.text())
        except ValueError:
            QMessageBox.warning(self, 'Invalid Input', 'Please enter valid numeric values for the Beam parameters.')
            return None

        # Swap if needed
        if target_bmaj < target_bmin:
            target_bmaj, target_bmin = target_bmin, target_bmaj
            self.bmaj_target.setText(f"{target_bmaj:.3g}")
            self.bmin_target.setText(f"{target_bmin:.3g}")

        self._last_action_name = "apply_smoothing_to_resolution"
        self._last_action_params = {
            "target_bmaj": float(target_bmaj),
            "target_bmin": float(target_bmin),
            "target_bpa": float(target_bpa),
            "current_bmaj": float(current_bmaj),
            "current_bmin": float(current_bmin),
            "current_bpa": float(current_bpa),
        }

        current_time = time.time()
        print("Convolving kernel...")

        try:
            smoothed_data, new_beam = usecases.compute_smoothed_to_resolution(
                self.original_data,
                self.header,
                target_bmaj=target_bmaj,
                target_bmin=target_bmin,
                target_bpa=target_bpa,
                current_bmaj=current_bmaj,
                current_bmin=current_bmin,
                current_bpa=current_bpa
            )
        except ValueError as e:
            QMessageBox.warning(self, 'Invalid Input', str(e))
            return None

        elapsed_time = time.time() - current_time
        print(f"Smoothing done in {elapsed_time:.3g} sec")

        # Store new beam parameters for save_fits
        self.new_bmaj = new_beam['BMAJ']
        self.new_bmin = new_beam['BMIN']
        self.new_bpa = new_beam['BPA']

        # Update header
        self.header['BMAJ'] = float(self.new_bmaj)
        self.header['BMIN'] = float(self.new_bmin)
        self.header['BPA'] = float(self.new_bpa)

        # Log kernel info
        FWHM_to_sigma = 1.0 / (2.0 * np.sqrt(2.0 * np.log(2)))
        pixel_scale_x = abs(self.header['CDELT1']) * 3600
        pixel_scale_y = abs(self.header['CDELT2']) * 3600

        sigma_kernel_maj_sq = (target_bmaj * FWHM_to_sigma) ** 2 - (current_bmaj * FWHM_to_sigma) ** 2
        sigma_kernel_min_sq = (target_bmin * FWHM_to_sigma) ** 2 - (current_bmin * FWHM_to_sigma) ** 2
        sigma_kernel_x = np.sqrt(sigma_kernel_min_sq) / pixel_scale_x
        sigma_kernel_y = np.sqrt(sigma_kernel_maj_sq) / pixel_scale_y
        theta = target_bpa - current_bpa

        fwhm_kernel_x = sigma_kernel_x / FWHM_to_sigma
        fwhm_kernel_y = sigma_kernel_y / FWHM_to_sigma
        print(f"\n\nCalculated Gaussian kernel size (FWHM): ({fwhm_kernel_y:.3g}, {fwhm_kernel_x:.3g}) pixels")
        print(f"Kernel rotation angle: {theta:.3g} deg")

        return smoothed_data

    def reset_smoothing(self):
        preferred_cursor = self._capture_preferred_cursor_snapshot()
        removed_preview = self._clear_preview_action()
        restored_from_history = False
        if removed_preview:
            restored_from_history = self._restore_state_from_action_history(preferred_cursor=preferred_cursor)

        # Fallback path: reset to the state when the panel was opened.
        if not restored_from_history and self.original_data is not None:
            self.fits_viewer.data = self.original_data.copy()
            self.fits_viewer.update_cube()
            for window in self.subwindows:
                if window:
                    window.data = self.fits_viewer.data
                    window.update_cube()

            if self.bmaj_exists:
                self.header['BMAJ'] = self.original_bmaj
            if self.bmin_exists:
                self.header['BMIN'] = self.original_bmin
            if self.bpa_exists:
                self.header['BPA'] = self.original_bpa

            if not self.bmaj_exists and 'BMAJ' in self.header:
                del self.header['BMAJ']
            if not self.bmin_exists and 'BMIN' in self.header:
                del self.header['BMIN']
            if not self.bpa_exists and 'BPA' in self.header:
                del self.header['BPA']

            self._sync_app_state_data()
            self.update_all_displays()
            self._restore_preferred_cursor_snapshot(preferred_cursor)

        # Keep panel references synchronized with the latest viewer state.
        self.data = self.fits_viewer.data
        self.header = self.fits_viewer.header
        self.wcs = self.fits_viewer.wcs
        self._refresh_current_resolution_display()
        self._refresh_main_hpbw_overlay()

        self._has_pending_changes = False
        self._last_action_name = None
        self._last_action_params = None
        
        print("\n\nReset")

    def _record_preview_action(self):
        if not self._last_action_name or not isinstance(self._last_action_params, dict):
            return
        record_action_preview(
            self.fits_viewer,
            self._last_action_name,
            dict(self._last_action_params),
            replace_tag=self._action_record_tag,
        )

    def _clear_preview_action(self):
        removed = clear_action_preview_record(
            self.fits_viewer,
            self._action_record_tag,
            action_name="apply_smoothing",
        )
        if removed:
            return True
        return bool(
            clear_action_preview_record(
                self.fits_viewer,
                self._action_record_tag,
                action_name="apply_smoothing_to_resolution",
            )
        )

    def _capture_preferred_cursor_snapshot(self):
        main_window = getattr(self.fits_viewer, 'main_window', None) or self.fits_viewer
        capture = getattr(main_window, "_capture_shared_cursor_snapshot", None)
        if callable(capture):
            try:
                snapshot = capture()
                if isinstance(snapshot, dict) and snapshot:
                    return dict(snapshot)
            except Exception:
                pass
        try:
            return {"zpix": int(self.fits_viewer.current_channel_index())}
        except Exception:
            return None

    def _restore_preferred_cursor_snapshot(self, snapshot) -> None:
        if not isinstance(snapshot, dict) or not snapshot:
            return
        main_window = getattr(self.fits_viewer, 'main_window', None) or self.fits_viewer
        zpix = snapshot.get("zpix")
        if zpix is None:
            return
        try:
            target_z = int(round(float(zpix)))
        except Exception:
            return
        update_channel = getattr(main_window, "update_channel", None)
        if callable(update_channel):
            try:
                update_channel("xy", target_z)
            except Exception:
                pass

    def _restore_state_from_action_history(self, *, preferred_cursor=None) -> bool:
        main_window = getattr(self.fits_viewer, 'main_window', None) or self.fits_viewer
        session = getattr(main_window, 'action_session', None)
        if session is None:
            return False
        replay_to_cursor = getattr(session, "_replay_to_cursor", None)
        apply_state = getattr(main_window, "_apply_action_session_state_to_viewers", None)
        if not callable(replay_to_cursor) or not callable(apply_state):
            return False
        try:
            history = list(getattr(session, "history", []) or [])
            cursor = int(getattr(session, "cursor", len(history)))
            cursor = max(0, min(cursor, len(history)))
        except Exception:
            return False
        try:
            replay_to_cursor(cursor)
            if isinstance(preferred_cursor, dict) and preferred_cursor:
                apply_state(preferred_cursor=dict(preferred_cursor))
            else:
                apply_state()
            return True
        except Exception:
            return False

    def _refresh_current_resolution_display(self):
        header = getattr(self, "header", None)
        if header is None:
            return
        try:
            if 'BMAJ' in header:
                self.bmaj_current.setText(f"{float(header['BMAJ']) * 3600:.3g}")
            else:
                self.bmaj_current.clear()
        except Exception:
            pass
        try:
            if 'BMIN' in header:
                self.bmin_current.setText(f"{float(header['BMIN']) * 3600:.3g}")
            else:
                self.bmin_current.clear()
        except Exception:
            pass
        try:
            if 'BPA' in header:
                self.bpa_current.setText(f"{float(header['BPA']):.3g}")
            else:
                self.bpa_current.clear()
        except Exception:
            pass

    def _sync_app_state_data(self):
        """Keep MainWindow.app_state data in sync with current cube after smoothing."""
        main_window = getattr(self.fits_viewer, 'main_window', None) or self.fits_viewer
        if hasattr(main_window, 'sync_app_state_data'):
            main_window.sync_app_state_data(
                data=self.fits_viewer.data,
                header=self.fits_viewer.header,
                wcs=self.fits_viewer.wcs,
            )
            return
        app_state = getattr(main_window, 'app_state', None)
        if app_state is not None:
            app_state.data = self.fits_viewer.data
            app_state.header = self.fits_viewer.header
            app_state.wcs = self.fits_viewer.wcs

    def _refresh_main_hpbw_overlay(self):
        hpbw = getattr(self.fits_viewer, 'hpbw', None)
        if hpbw is None:
            return
        try:
            hpbw.header = self.fits_viewer.header
        except Exception:
            pass
        refresh = getattr(hpbw, 'refresh_geometry_from_header', None)
        if callable(refresh):
            try:
                refresh()
            except Exception:
                pass
        redraw_overlay = getattr(self.fits_viewer, 'redraw_main_overlay_and_blit', None)
        if callable(redraw_overlay):
            try:
                redraw_overlay()
                return
            except Exception:
                pass
        canvas = getattr(self.fits_viewer, 'canvas', None)
        if canvas is not None:
            try:
                canvas.draw_idle()
            except Exception:
                pass
            

    def update_all_displays(self):
        # Get the current channel index for the main viewer
        current_channel_main = self.fits_viewer.current_channel_index()
    
        # Update the display for the main viewer
        if self.fits_viewer.data.ndim == 4:
            if self.fits_viewer.plane == 'xy':
                data_slice = self.fits_viewer.data[0, current_channel_main, :, :] 
            elif self.fits_viewer.plane == 'xz':
                data_slice = self.fits_viewer.data[0, :, current_channel_main, :]
            elif self.fits_viewer.plane == 'zy':
                data_slice = self.fits_viewer.data[0, :, :, current_channel_main].T
            else:
                data_slice = self.fits_viewer.data[0, 0]  
    
        elif self.fits_viewer.data.ndim == 3:
            if self.fits_viewer.plane == 'xy':
                data_slice = self.fits_viewer.data[current_channel_main, :, :]
            elif self.fits_viewer.plane == 'xz':
                data_slice = self.fits_viewer.data[:, current_channel_main, :]
            elif self.fits_viewer.plane == 'zy':
                data_slice = self.fits_viewer.data[:, :, current_channel_main].T
            else:
                data_slice = self.fits_viewer.data[0]
    
        elif self.fits_viewer.data.ndim == 2:
            data_slice = self.fits_viewer.data
        else:
            raise TypeError(f"Invalid number of dimensions: {self.fits_viewer.data.ndim}")
    
        # Update the image data and color limits
        if data_slice.ndim == 2:
            self.fits_viewer.im.set_data(data_slice)
        else:
            raise TypeError(f"Invalid data shape {data_slice.shape} for 2D image display.")
    
        self.fits_viewer.canvas.draw()
    
        # Update the displays for all subwindows
        for window in self.subwindows:
            if window:
                current_channel = window.current_channel_index()
                if window.data.ndim == 4:
                    if window.plane == 'xy':
                        data_slice = window.data[0, current_channel, :, :]
                    elif window.plane == 'xz':
                        data_slice = window.data[0, :, current_channel, :]
                    elif window.plane == 'zy':
                        data_slice = window.data[0, :, :, current_channel].T
                    else:
                        data_slice = window.data[0, 0]
                        
                elif window.data.ndim == 3:
                    if window.plane == 'xy':
                        data_slice = window.data[current_channel, :, :]
                    elif window.plane == 'xz':
                        data_slice = window.data[:, current_channel, :]
                    elif window.plane == 'zy':
                        data_slice = window.data[:, :, current_channel].T
                    else:
                        data_slice = window.data[0]
    
                elif window.data.ndim == 2:
                    data_slice = window.data
                else:
                    raise TypeError(f"Invalid number of dimensions: {window.data.ndim}")
    
                if data_slice.ndim == 2:
                    window.im.set_data(data_slice)
                else:
                    raise TypeError(f"Invalid data shape {data_slice.shape} for 2D image display.")
                    
                window.canvas.draw()

    def save_fits(self):
        self.execute_smoothing()
        data_min =  float(np.nanmin(self.fits_viewer.data))
        data_max =  float(np.nanmax(self.fits_viewer.data))
        new_header = self.fits_viewer.header.copy()
        original_filename = self.fits_viewer.filename
        
        new_header['DATAMIN'] = data_min
        new_header['DATAMAX'] = data_max

        if self.checkbox.isChecked():
            new_header['BMAJ'] = float(self.new_bmaj)
            new_header['BMIN'] = float(self.new_bmin)
            new_header['BPA'] = float(self.new_bpa)

        # Use common centralized processing history

        for entry in build_processing_history_lines(self.fits_viewer):
            new_header.add_history(entry)

        save_fits = SaveFITS(self.fits_viewer.data, new_header, original_filename)
        save_fits.save(suffix="sm")

    def move_to_default_position(self):
        # Get MainWindow geometry
        mainwindow_geometry = self.fits_viewer.geometry()
        mainwindow_x = mainwindow_geometry.x()
        mainwindow_y = mainwindow_geometry.y()
        mainwindow_width = mainwindow_geometry.width()

        # Move ControlPanel to the right of MainWindow
        self.move(mainwindow_x + mainwindow_width, mainwindow_y - 28)

    def closeEvent(self, event):
        if self._has_pending_changes:
            choice = confirm_pending_close(
                self,
                "Close Smoothing Panel",
                "There are unapplied smoothing changes.",
            )
            if choice == "cancel":
                event.ignore()
                return
            if choice == "discard":
                self.reset_smoothing()
        super().closeEvent(event)

    def resync_after_workspace_restore(self):
        self._has_pending_changes = bool(has_action_record_tag(self.fits_viewer, self._action_record_tag))



from PySide6.QtWidgets import QDialog, QVBoxLayout, QTextEdit, QDialogButtonBox
from PySide6.QtGui import QFont
from PySide6.QtCore import Qt

class MomentResultsDialog(QDialog):
    """
    A non-modal dialog to display text results in a selectable text box.
    It deletes itself upon closing to prevent memory leaks.
    """
    def __init__(self, title, content, parent=None):
        super().__init__(parent)
        self.setWindowTitle(title)
        # Ensure the dialog is deleted when closed
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)

        # Main layout
        layout = QVBoxLayout(self)

        # Text edit area to display the content
        self.text_edit = QTextEdit()
        self.text_edit.setReadOnly(True)
        self.text_edit.setText(content)
        # Use a standard system font for a cleaner look
        self.text_edit.setFont(QFont())
        
        layout.addWidget(self.text_edit)

        # OK button to close the dialog
        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok)
        button_box.accepted.connect(self.accept)
        layout.addWidget(button_box)

        self.resize(500, 400)


"""Shared helper functions used by multiple tool panels."""
from __future__ import annotations


def _resolve_xz_subwindow(subwindows):
    """Return the first (XZ) subwindow, or None if unavailable."""
    if not subwindows:
        return None
    try:
        window = subwindows[0]
    except Exception:
        return None
    return window if window is not None else None


def _resolve_z_view_limits(fits_viewer, subwindows):
    """Return the Z-axis pixel limits ``(zmin, zmax)`` from the best available source."""
    xz_window = _resolve_xz_subwindow(subwindows)
    if xz_window is not None:
        try:
            return tuple(xz_window.ax.get_ylim())
        except Exception:
            pass

    limits = getattr(fits_viewer, "original_zlim", None)
    if isinstance(limits, (tuple, list)) and len(limits) == 2:
        return tuple(limits)

    data = getattr(fits_viewer, "data", None)
    if getattr(data, "ndim", 0) > 2:
        depth_len = data.shape[data.ndim - 3]
        return (-0.5, depth_len - 0.5)

    return (0.0, 0.0)



"""
Color mode enumeration for view settings.

This module defines ColorMode which specifies which view plane's color settings
are being configured. This is placed in core/ to avoid layer violations (ui/tools
importing from each other).
"""
from enum import Enum


class ColorMode(Enum):
    """
    Enumeration for different color mode contexts.
    
    - MAIN: Main viewer (xy plane)
    - INTEG: Integration result window
    - PV: PV diagram window
    - CHANNEL: Channel map window
    """
    MAIN = "main"
    INTEG = "integ"
    PV = "pv"
    CHANNEL = "channel"


def default_color_settings() -> dict:
    return {
        "min_val": None,
        "max_val": None,
        "log_scale": False,
        "gamma_value": 1.0,
        "invert": False,
        "color_pattern": None,
    }


def default_color_settings_map() -> dict:
    return {mode: default_color_settings() for mode in ColorMode}



"""Clump finding usecases."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import numpy as np

from takefits.core.app_state import AppState
from .utils import update_datamin_datamax_if_present


@dataclass
class ClumpResult:
    """Result from clump finding algorithm."""
    mask: np.ndarray  # Integer label mask
    n_clumps: int
    catalog: list  # List of dicts with properties per clump
    algorithm: str
    parameters: Dict[str, Any]


def run_clumpfind(
    state: AppState,
    rms: float,
    min_threshold_sigma: float = 3.0,
    step_sigma: float = 2.0,
    min_pixels: int = 10
) -> ClumpResult:
    """
    Run ClumpFind algorithm on state data.

    Args:
        state: AppState with data
        rms: RMS noise level (in data units)
        min_threshold_sigma: Minimum threshold in units of RMS
        step_sigma: Contour step size in units of RMS
        min_pixels: Minimum pixels per clump

    Returns:
        ClumpResult with mask and catalog
    """
    from takefits.logic.clumpfind import ClumpFind

    if state.data is None:
        raise ValueError("No data loaded")

    data = state.data
    if data.ndim == 4:
        data = data[state.current_s]

    min_val = rms * min_threshold_sigma
    step = rms * step_sigma

    finder = ClumpFind(data, wcs=state.wcs)
    mask = finder.run(min_val=min_val, step=step, min_pix=min_pixels)
    catalog = finder.get_catalog()

    n_clumps = len(np.unique(mask)) - 1  # Exclude 0 (background)
    if n_clumps < 0:
        n_clumps = 0

    return ClumpResult(
        mask=mask,
        n_clumps=n_clumps,
        catalog=catalog,
        algorithm="clumpfind",
        parameters={
            "rms": rms,
            "min_threshold_sigma": min_threshold_sigma,
            "step_sigma": step_sigma,
            "min_pixels": min_pixels,
            "min_val": min_val,
            "step": step
        }
    )


def run_fellwalker(
    state: AppState,
    rms: float,
    min_threshold_sigma: float = 3.0,
    min_dip_sigma: float = 2.0,
    min_pixels: int = 10
) -> ClumpResult:
    """
    Run FellWalker (watershed-based) algorithm.

    Args:
        state: AppState with data
        rms: RMS noise level (in data units)
        min_threshold_sigma: Minimum threshold in units of RMS
        min_dip_sigma: Minimum dip (prominence) in units of RMS
        min_pixels: Minimum pixels per clump

    Returns:
        ClumpResult with mask and catalog
    """
    from takefits.logic.fellwalker import FellWalker

    if state.data is None:
        raise ValueError("No data loaded")

    data = state.data
    if data.ndim == 4:
        data = data[state.current_s]

    min_val = rms * min_threshold_sigma
    min_dip = rms * min_dip_sigma

    walker = FellWalker(data, wcs=state.wcs)
    mask = walker.run(min_val=min_val, min_dip=min_dip, min_pix=min_pixels)
    catalog = walker.get_catalog()

    n_clumps = len(np.unique(mask)) - 1  # Exclude 0 (background)
    if n_clumps < 0:
        n_clumps = 0

    return ClumpResult(
        mask=mask,
        n_clumps=n_clumps,
        catalog=catalog,
        algorithm="fellwalker",
        parameters={
            "rms": rms,
            "min_threshold_sigma": min_threshold_sigma,
            "min_dip_sigma": min_dip_sigma,
            "min_pixels": min_pixels,
            "min_val": min_val,
            "min_dip": min_dip
        }
    )


def run_dendrogram(
    state: AppState,
    rms: float,
    min_value_sigma: float = 3.0,
    min_delta_sigma: float = 2.0,
    min_npix: int = 10,
    output_mode: str = "leaves",
    use_scimes: bool = False,
    scimes_criteria: Optional[list] = None,
    scimes_user_k: int = 0,
    scimes_save_isol: bool = True
) -> ClumpResult:
    """
    Run Dendrogram algorithm (optionally with SCIMES clustering).

    Args:
        state: AppState with data
        rms: RMS noise level (in data units)
        min_value_sigma: Minimum value in units of RMS
        min_delta_sigma: Minimum delta (height difference) in units of RMS
        min_npix: Minimum pixels per structure
        output_mode: "leaves", "roots", or "all"
        use_scimes: Whether to apply SCIMES clustering
        scimes_criteria: List of criteria for SCIMES (e.g., ["luminosity", "volume"])
        scimes_user_k: Target number of clusters for SCIMES (0=auto)
        scimes_save_isol: Whether to include isolated leaves in SCIMES result

    Returns:
        ClumpResult with mask and catalog
    """
    from takefits.logic.dendro_handler import DendroHandler

    if state.data is None:
        raise ValueError("No data loaded")

    data = state.data
    if data.ndim == 4:
        data = data[state.current_s]

    min_value = rms * min_value_sigma
    min_delta = rms * min_delta_sigma

    # Pass header to handler!
    handler = DendroHandler(data, wcs=state.wcs, header=state.header)
    handler.run_dendrogram(min_value=min_value, min_delta=min_delta, min_npix=min_npix)

    if use_scimes and scimes_criteria:
        # Pass rms and save_isol_leaves
        handler.run_scimes(
            criteria=scimes_criteria,
            user_k=scimes_user_k,
            rms=rms,
            save_isol_leaves=scimes_save_isol
        )

    mask = handler.get_mask(mode=output_mode)
    catalog = handler.get_catalog()

    n_clumps = len(np.unique(mask)) - 1  # Exclude 0 (background)
    if n_clumps < 0:
        n_clumps = 0

    parameters = {
        "rms": rms,
        "min_value_sigma": min_value_sigma,
        "min_delta_sigma": min_delta_sigma,
        "min_npix": min_npix,
        "output_mode": output_mode,
        "use_scimes": use_scimes,
        "min_value": min_value,
        "min_delta": min_delta
    }

    if use_scimes:
        parameters.update({
            "scimes_criteria": scimes_criteria,
            "scimes_user_k": scimes_user_k,
            "scimes_save_isol": scimes_save_isol,
        })

    return ClumpResult(
        mask=mask,
        n_clumps=n_clumps,
        catalog=catalog,
        algorithm="dendrogram",
        parameters=parameters
    )


def check_scimes_availability() -> tuple[bool, str]:
    """
    Check if SCIMES is available.

    Returns:
        (True, "") if SCIMES is available.
        (False, error_message) if it is not.
    """
    try:
        from takefits.logic.dendro_handler import DendroHandler
        try:
            import takefits.logic.scimes
            return True, ""
        except ImportError as e:
            return False, f"SCIMES module missing dependency: {e}"
    except ImportError as e:
        return False, f"astrodendro or other dendro dependency missing: {e}"


def generate_catalog(state: AppState, mask: np.ndarray) -> List[Dict[str, Any]]:
    """
    Generate a catalog of properties for clumps in the provided mask.

    Args:
        state: AppState containing data and WCS
        mask: Label mask (same shape as state.data)

    Returns:
        List of property dictionaries (one per clump)
    """
    from takefits.logic.cloud_catalog_utils import calculate_moments_and_props

    if state.data is None:
        raise ValueError("No data loaded")

    # Use data from state (handling 4th axis if needed)
    data = state.data
    if data.ndim == 4:
        data = data[state.current_s]

    if mask.shape != data.shape:
        raise ValueError(f"Mask shape {mask.shape} does not match data shape {data.shape}")

    labels = np.unique(mask)
    labels = labels[labels > 0]

    catalog = []
    for l in labels:
        props = calculate_moments_and_props(data, mask, l, wcs=state.wcs)
        if props:
            catalog.append(props)

    return catalog


def export_clump_mask(
    state: AppState,
    result: ClumpResult,
    output_path: str,
    history_entries: Optional[list] = None
) -> str:
    """
    Export clump mask to FITS file.

    Args:
        state: AppState with original header info
        result: ClumpResult from clump finding
        output_path: Path for output FITS file
        history_entries: Optional list of HISTORY entries

    Returns:
        The output file path
    """
    from astropy.io import fits

    header = state.header.copy() if state.header is not None else fits.Header()

    # Update shape info
    header['NAXIS'] = result.mask.ndim
    for i, dim in enumerate(result.mask.shape[::-1], 1):
        header[f'NAXIS{i}'] = dim

    # Add clump finding info
    header['CLUMPALG'] = (result.algorithm, 'Clump finding algorithm')
    header['NCLUMPS'] = (result.n_clumps, 'Number of clumps found')

    # Add algorithm parameters
    for key, value in result.parameters.items():
        if isinstance(value, (int, float, str, bool)):
            key_fits = key[:8].upper()  # FITS keyword limit
            header[key_fits] = value

    # Add history
    if history_entries:
        for entry in history_entries:
            header.add_history(entry)

    update_datamin_datamax_if_present(header, result.mask)

    # Write file
    hdu = fits.PrimaryHDU(data=result.mask.astype(np.int32), header=header)
    hdu.writeto(output_path, overwrite=True)

    return output_path


def export_clump_catalog(
    result: ClumpResult,
    output_path: str,
    format: str = "csv"
) -> str:
    """
    Export clump catalog to CSV or FITS table.

    Args:
        result: ClumpResult from clump finding
        output_path: Path for output file
        format: "csv" or "fits"

    Returns:
        The output file path
    """
    if not result.catalog:
        # Write empty file
        if format == "csv":
            with open(output_path, 'w') as f:
                f.write("# No clumps found\n")
        else:
            from astropy.io import fits
            hdu = fits.BinTableHDU.from_columns([])
            hdu.writeto(output_path, overwrite=True)
        return output_path

    if format == "csv":
        import csv

        # Get all keys from catalog entries
        all_keys = set()
        for entry in result.catalog:
            all_keys.update(entry.keys())
        all_keys = sorted(all_keys)

        with open(output_path, 'w', newline='') as f:
            writer = csv.DictWriter(f, fieldnames=all_keys)
            writer.writeheader()
            for entry in result.catalog:
                writer.writerow(entry)

    elif format == "fits":
        from astropy.io import fits
        from astropy.table import Table

        # Convert catalog to astropy Table
        table = Table(result.catalog)
        table.write(output_path, format='fits', overwrite=True)

    else:
        raise ValueError(f"Unknown format: {format}")

    return output_path



import numpy as np
from scipy import ndimage
import sys

class ClumpFind:
    """
    Custom implementation of Clumpfind (Williams et al. 1994).
    """
    def __init__(self, data, wcs=None):
        self.data = data
        self.wcs = wcs
        self.clump_mask = np.zeros_like(self.data, dtype=int)
        self.clump_id_counter = 0

    def run(self, min_val, step, min_pix=5):
        """
        Run the Clumpfind algorithm.

        Args:
            min_val (float): Integration or Intensity threshold to start (lowest level).
            step (float): Step size for contour levels.
            min_pix (int): Minimum number of pixels for a valid clump.

        Returns:
            np.ndarray: Label mask where each clump has a unique integer ID.
        """
        # 1. Define Contour Levels
        # Generate levels from max down to min_val
        max_val = np.nanmax(self.data)

        if np.isnan(max_val) or max_val < min_val:
            print(f"Invalid max_val: {max_val}. Lower than min_val {min_val} or NaN.")
            return np.zeros_like(self.data, dtype=int)

        # Safety Check for array size
        # If step is too small, levels array explodes
        range_val = max_val - min_val
        n_steps = range_val / step
        if n_steps > 10000:
            print(f"Warning: Too many steps ({n_steps}). Clamping to 1000 steps.")
            step = range_val / 1000.0

        levels = np.arange(min_val, max_val + step, step)
        levels = levels[::-1] # Check from highest to lowest

        self.clump_mask = np.zeros_like(self.data, dtype=int)
        self.clump_id_counter = 0
        
        n_levels = len(levels)
        print(f"ClumpFind: Processing {n_levels} contour levels...")

        # 2. Iterate through levels
        for i, level in enumerate(levels):
            # Progress update
            if i % 5 == 0 or i == n_levels - 1:
                progress = (i + 1) / n_levels * 100
                sys.stdout.write(f"\rClumpFind Progress: {progress:.1f}% (Level {i+1}/{n_levels})")
                sys.stdout.flush()

            # Create a binary mask for the current level
            binary_mask = self.data > level

            # Label connected components at this level
            # structure=np.ones((3,3)) defines connectivity (diagonals included for 2D)
            # For Clumpfind, usually we want strict connectivity (4 for 2D, 6 for 3D) to keep peaks separate
            # so we use generate_binary_structure(ndim, 1)
            ndim = self.data.ndim
            structure = ndimage.generate_binary_structure(ndim, 1) # 1-connectivity (cross)

            # Regions above current level
            labeled_regs, num_regs = ndimage.label(binary_mask, structure=structure)

            if num_regs == 0:
                continue

            # Analyze each connected region at this level
            slices = ndimage.find_objects(labeled_regs)

            for i in range(num_regs):
                label = i + 1
                region_slice = slices[i]

                # Extract the mask for this specific region
                # This is a boolean mask of the current region within its bounding box (roughly)
                # But to access data we should use the full mask or careful indexing
                # Let's interact with the full clump_mask

                # Extract sub-arrays for faster processing
                # We work within the bounding box (region_slice) to avoid full-size mask operations
                sub_labeled = labeled_regs[region_slice]
                sub_region_mask = (sub_labeled == label)
                sub_clump_mask = self.clump_mask[region_slice]

                # Find IDs of existing clumps that overlap with this new region (within the slice)
                existing_labels = sub_clump_mask[sub_region_mask]
                existing_ids = np.unique(existing_labels)
                existing_ids = existing_ids[existing_ids > 0] # Ignore 0

                if len(existing_ids) == 0:
                    # Case 1: No overlap -> New Clump
                    self.clump_id_counter += 1
                    # Update only within the slice using the sub_region_mask
                    # sub_clump_mask is a view, so this updates the main array
                    sub_clump_mask[sub_region_mask] = self.clump_id_counter

                elif len(existing_ids) == 1:
                    # Case 2: One overlap -> Extension
                    clump_id = existing_ids[0]
                    sub_clump_mask[sub_region_mask] = clump_id

                else:
                    # Case 3: Merge (Friends-of-Friends)
                    # We grow existing labels into unassigned pixels

                    # Unassigned mask WITHIN THE SLICE
                    sub_unassigned = sub_region_mask & (sub_clump_mask == 0)
                    if not np.any(sub_unassigned):
                        continue

                    # We need a working copy of labels in this slice
                    current_labels = sub_clump_mask.copy()

                    # We only care about growing into 'sub_region_mask'
                    # Iterative dilation
                    # For strict connectivity:
                    structure = ndimage.generate_binary_structure(ndim, 1)

                    # Mask of pixels we want to fill (valid region but currently 0)
                    # This is essentially sub_unassigned, but let's be explicit
                    target_mask = sub_unassigned & (current_labels == 0)

                    while np.any(target_mask):
                        # Dilate the current non-zero labels
                        # We use grey_dilation or maximum_filter to propagate labels?
                        # maximum_filter propagates the largest label. This might bias towards larger IDs if they touch?
                        # Williams: "pixels are added to the clumps they touch".
                        # If a pixel touches label 1 and label 2? Conflict handling.
                        # Standard FoF usually creates a Voronoi-like partition.
                        # maximum_filter is a reasonable approximation for simultaneous growth.

                        dilated = ndimage.maximum_filter(current_labels, footprint=structure)

                        # Identify new pixels to fill:
                        # Must be in target_mask AND have been touched by dilation (dilated > 0)
                        newly_filled = (dilated > 0) & target_mask

                        if not np.any(newly_filled):
                            # No progress - maybe disconnected components in target_mask that can't be reached?
                            # This shouldn't happen if region is connected, unless seeds are disjoint from targets within this region?
                            # But 'region' comes from label(), so it is connected.
                            # 'seeds' and 'targets' are both parts of this connected region.
                            break

                        # Update labels
                        current_labels[newly_filled] = dilated[newly_filled]

                        # Update target mask (remove filled pixels)
                        target_mask[newly_filled] = False

                    # Write back to main mask
                    # Only update the pixels that were originally unassigned
                    # current_labels now has the filled values

                    # The 'unassigned_mask' corresponds to where we wanted to fill.
                    # We can just write back the changed parts.

                    # sub_unassigned is a boolean mask of the SLICE
                    # current_labels corresponds to the SLICE

                    # We need to act on the main self.clump_mask using the SLICE + MASK

                    # Get the specific pixels in the slice that were updated
                    update_values = current_labels[sub_unassigned]

                    # Apply to main mask
                    # We need to access self.clump_mask[region_slice] and update specific pixels
                    # self.clump_mask[region_slice] is a view?
                    # Yes, basic slicing returns a view.

                    # However, applying a boolean mask 'sub_unassigned' to that view might trigger a copy if not done carefully.
                    # self.clump_mask[region_slice][sub_unassigned] = ... works if __setitem__ handles it.

                    # Safer way:
                    slice_view = self.clump_mask[region_slice]
                    slice_view[sub_unassigned] = update_values


        # 3. Post-processing
        # Remove small clumps (min_pix)
        if min_pix > 0:
            print(f"ClumpFind: Filtering small clumps (<{min_pix} pix)...")
            # Vectorized size filtering
            # 1. Count pixels per label
            # bincount works on 1D array of non-negative integers
            counts = np.bincount(self.clump_mask.ravel())
            
            # 2. Identify dense labels (size >= min_pix)
            # Create a mapping array: existing_label -> new_label (or 0 if filtered)
            # For ClumpFind, we just want to zero out small ones, preserving other IDs?
            # Or do we want to be safe? The original code just zeroed them out.
            
            # map_array[label] = label if size >= min_pix else 0
            map_array = np.arange(len(counts), dtype=self.clump_mask.dtype)
            map_array[counts < min_pix] = 0 # Zero out small clumps
            
            # 3. Apply mapping
            self.clump_mask = map_array[self.clump_mask]

        print("\nClumpFind: Finished.")
        return self.clump_mask

    def get_catalog(self):
        """
        Generate a catalog of properties for detected clumps.
        Requires self.clump_mask to be populated.
        """
        if self.clump_mask is None:
            return []

        labels = np.unique(self.clump_mask)
        labels = labels[labels > 0]

        from takefits.logic.cloud_catalog_utils import calculate_moments_and_props

        catalog = []
        for l in labels:
            props = calculate_moments_and_props(self.data, self.clump_mask, l, self.wcs)
            if props:
                catalog.append(props)

        return catalog



import numpy as np

def calculate_props_from_indices_values(indices, values, wcs=None, ndim=3, label_id=0):
    """
    Calculate properties from indices and values.
    shared logic for ClumpFind, FellWalker, and Dendrogram.
    """
    if len(values) == 0:
        return None

    # 0. Basic Stats
    npix = len(values)
    flux_sum = np.sum(values)
    peak_val = np.max(values)
    mean_val = np.mean(values)

    # Weights for moments (Intensity)
    weights = values
    sum_w = np.sum(weights)

    if sum_w == 0: sum_w = 1.0 # Avoid division by zero

    # 1. Centroids (Weighted)
    # indices is tuple of arrays 
    # ndim should match len(indices)

    coords_pix = []

    # Calculate simple weighted mean for each axis
    centroids_pix = []
    for i in range(ndim):
        # Ensure indices[i] exists
        if i < len(indices):
            c = np.sum(indices[i] * weights) / sum_w
            centroids_pix.append(c)
        else:
            centroids_pix.append(0.0)

    # Assign to standardized variable names based on dimensionality
    if ndim == 3:
        z_c_pix, y_c_pix, x_c_pix = centroids_pix
    elif ndim == 2:
        y_c_pix, x_c_pix = centroids_pix
        z_c_pix = 0
    else:
        x_c_pix = centroids_pix[0]
        y_c_pix, z_c_pix = 0, 0

    # 2. Second Moments (Dispersion)
    # Sigma^2 = Sum( w * (x - x_c)^2 ) / Sum(w)

    sigmas_pix = []
    
    # Store per-axis sigma for generic access
    sigma_dict = {}

    for i in range(ndim):
        if i < len(indices):
            diff = indices[i] - centroids_pix[i]
            sigma2 = np.sum(weights * (diff**2)) / sum_w
            sigma = np.sqrt(sigma2)
            sigmas_pix.append(sigma)
        else:
            sigmas_pix.append(0.0)

    if ndim == 3:
        z_sig_pix, y_sig_pix, x_sig_pix = sigmas_pix
        sigma_dict['z'] = z_sig_pix
        sigma_dict['y'] = y_sig_pix
        sigma_dict['x'] = x_sig_pix
    elif ndim == 2:
        y_sig_pix, x_sig_pix = sigmas_pix
        z_sig_pix = 0
        sigma_dict['y'] = y_sig_pix
        sigma_dict['x'] = x_sig_pix
    else:
        x_sig_pix = sigmas_pix[0]
        y_sig_pix, z_sig_pix = 0, 0
        sigma_dict['x'] = x_sig_pix

    # 3. Covariance for Ellipse (Spatial: X vs Y)
    # Map indices to X, Y
    # dim 2 (X) corresponds to index -1
    # dim 1 (Y) corresponds to index -2
    
    if ndim >= 2:
        idx_x = indices[-1]
        idx_y = indices[-2]
        
        diff_x = idx_x - x_c_pix
        diff_y = idx_y - y_c_pix

        cov_xy = np.sum(weights * diff_x * diff_y) / sum_w
        var_x = x_sig_pix**2
        var_y = y_sig_pix**2

        term = np.sqrt((var_x - var_y)**2 + 4*cov_xy**2)
        lambda1 = (var_x + var_y + term) / 2
        lambda2 = (var_x + var_y - term) / 2

        maj_sigma_pix = np.sqrt(lambda1)
        min_sigma_pix = np.sqrt(lambda2)
        
        # Standard: CCW from positive X.
        pa_rad = 0.5 * np.arctan2(2*cov_xy, var_x - var_y)
        pa_deg = np.degrees(pa_rad)
    else:
        maj_sigma_pix = 0
        min_sigma_pix = 0
        pa_deg = 0

    # 4. Physical Units (WCS)
    sigma_x_deg = 0
    sigma_y_deg = 0
    sigma_v_kms = 0
    maj_sigma_deg = 0
    min_sigma_deg = 0
    r_eff_deg = 0
    area_deg2 = 0

    glon_deg, glat_deg, velo_kms = 0, 0, 0

    if wcs:
        from astropy.wcs.utils import proj_plane_pixel_scales
        try:
            scales = proj_plane_pixel_scales(wcs.celestial)
            dx_deg = scales[0] # Approx
            dy_deg = scales[1] if len(scales) > 1 else dx_deg

            # Area in deg2
            pixel_area_deg2 = dx_deg * dy_deg
            area_deg2 = npix * pixel_area_deg2

            # R_eff = sqrt(Area / pi)
            r_eff_deg = np.sqrt(area_deg2 / np.pi)

            # Sigmas
            sigma_x_deg = x_sig_pix * dx_deg
            sigma_y_deg = y_sig_pix * dy_deg

            maj_sigma_deg = maj_sigma_pix * max(dx_deg, dy_deg) 
            min_sigma_deg = min_sigma_pix * min(dx_deg, dy_deg)

        except Exception:
            # Fallback
            if hasattr(wcs.wcs, 'cdelt'):
                cdelt = wcs.wcs.cdelt
                dx_deg = abs(cdelt[0])
                dy_deg = abs(cdelt[1])
                area_deg2 = npix * dx_deg * dy_deg
                r_eff_deg = np.sqrt(area_deg2 / np.pi)
                sigma_x_deg = x_sig_pix * dx_deg
                sigma_y_deg = y_sig_pix * dy_deg
                maj_sigma_deg = maj_sigma_pix * dx_deg
                min_sigma_deg = min_sigma_pix * dy_deg

        
        # Velocity & World Coordinate Logic
        try:
            wcs_naxis = wcs.wcs.naxis if hasattr(wcs.wcs, 'naxis') else ndim
            
            # A. Velocity (Manual 3D Calculation)
            # User requested simple calculation from CDELT/CRVAL/CRPIX to avoid unit issues.
            if ndim == 3 and wcs_naxis >= 3:
                 try:
                     spec_axis = 2 # 3rd axis
                     cdelt = wcs.wcs.cdelt[spec_axis]
                     crval = wcs.wcs.crval[spec_axis]
                     crpix = wcs.wcs.crpix[spec_axis] # 1-based
                     
                     # Unit Heuristic: Check CUNIT or CDELT magnitude
                     # If CUNIT is 'm/s', convert to km/s
                     # If CUNIT is missing but CDELT is huge (> 100), assume m/s and convert
                     try:
                         pixel_scale = 1.0
                         cunit = wcs.wcs.cunit[spec_axis].to_string().strip().lower()
                         if cunit == 'm/s':
                             pixel_scale = 0.001
                         elif cunit == '' and abs(cdelt) > 50.0:
                             # Heuristic: Most radio data has channel width < 50 km/s. 
                             # If values are > 50, it is likely m/s (e.g. 3000 m/s).
                             pixel_scale = 0.001
                         
                         cdelt *= pixel_scale
                         crval *= pixel_scale
                     except Exception:
                         # Fallback heuristic
                         if abs(cdelt) > 50.0:
                             cdelt *= 0.001
                             crval *= 0.001

                     # 1. Sigma (Dispersion)
                     sigma_v_kms = z_sig_pix * abs(cdelt)
                     
                     # 2. Centroid
                     # z_c_pix is 0-based index
                     velo_kms = crval + (z_c_pix - (crpix - 1)) * cdelt
                 except Exception:
                     pass

            # B. Spatial World Coordinates (RA/DEC / GLON/GLAT)
            # Use pixel_to_world for spatial axes
            if wcs_naxis == 3:
                # 3D WCS
                try:
                    world = wcs.pixel_to_world(x_c_pix, y_c_pix, z_c_pix)
                    if isinstance(world, (list, tuple)):
                        coord = world[0]
                        # We used manual velocity above, so ignore world[1] or verify
                    else:
                        coord = world
                    
                    if hasattr(coord, 'l'):
                        glon_deg = coord.l.deg
                        glat_deg = coord.b.deg
                    elif hasattr(coord, 'ra'):
                        glon_deg = coord.ra.deg
                        glat_deg = coord.dec.deg
                except Exception:
                    pass
                    
            elif wcs_naxis == 2 or ndim == 2:
                # 2D WCS
                try:
                    try:
                        celestial_wcs = wcs.celestial
                        world = celestial_wcs.pixel_to_world(x_c_pix, y_c_pix)
                    except Exception:
                        world = wcs.pixel_to_world(x_c_pix, y_c_pix)

                    if hasattr(world, 'l'):
                        glon_deg = world.l.deg
                        glat_deg = world.b.deg
                    elif hasattr(world, 'ra'):
                        glon_deg = world.ra.deg
                        glat_deg = world.dec.deg
                except Exception:
                    pass

        except Exception:
            pass

    return {
        'id': int(label_id),
        'peak': peak_val,
        'mean': mean_val,
        'sum': flux_sum,
        'npix': npix,
        'x_pix': x_c_pix,
        'y_pix': y_c_pix,
        'z_pix': z_c_pix,
        'cen_x_deg': glon_deg,
        'cen_y_deg': glat_deg,
        'cen_v_kms': velo_kms,
        'sigma_x_pix': x_sig_pix,
        'sigma_y_pix': y_sig_pix,
        'sigma_z_pix': z_sig_pix,
        'sigma_x_deg': sigma_x_deg,
        'sigma_y_deg': sigma_y_deg,
        'sigma_v_kms': sigma_v_kms,
        'maj_sigma_deg': maj_sigma_deg,
        'min_sigma_deg': min_sigma_deg,
        'pa_deg': pa_deg,
        'area_deg2': area_deg2,
        'r_eff_deg': r_eff_deg
    }

def calculate_moments_and_props(data, mask, label, wcs=None):
    """
    Calculate properties for a specific label in the mask.
    Wrapper around calculate_props_from_indices_values.
    """
    obj_mask = (mask == label)
    if not np.any(obj_mask):
        return None

    indices = np.where(obj_mask)
    values = data[indices]

    return calculate_props_from_indices_values(indices, values, wcs=wcs, ndim=data.ndim, label_id=label)
